package marker;

import java.awt.Color;
import java.util.Collections;
import java.util.Vector;

import javax.swing.JComponent;

import marker.MarkerSetsPlugin.Event;

import org.gjt.sp.jedit.Buffer;
import org.gjt.sp.jedit.View;
import org.gjt.sp.jedit.jEdit;
import org.gjt.sp.jedit.gui.DockableWindowManager;
import org.gjt.sp.jedit.msg.BufferUpdate;
import org.gjt.sp.jedit.textarea.JEditTextArea;
import org.gjt.sp.jedit.visitors.JEditVisitorAdapter;
import org.gjt.sp.util.SyntaxUtilities;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/** A named container of FileMarker objects, implemented as a vector.
*/
public class MarkerSet {

	private static final String COLOR_ATTR = "color";
	private static final String NAME_ATTR = "name";
	private String name;
	private Vector<FileMarker> markers;
	private Color color;
	private int currentMarker;
	
	/** A named, empty MarkerSet. 
	    @param name the name of the MarkerSet.
	*/
	public MarkerSet(String name) {
		this.name = name;
		color = Color.black;
		markers = new Vector<FileMarker>();
		currentMarker = -1;
	}
	
	/** A MarkerSet de-serialized from an XML DOM Node. 
	    @param node the XML DOM node 
	*/
	public MarkerSet(Element node) {
		markers = new Vector<FileMarker>();
		importXml(node);
		currentMarker = -1;
	}
	
	public void setColor(Color c) { color = c; }
	public Color getColor() { return color; }
	
	public String getName() { return name; }

	public FileMarker nextMarker(View view) {
		int size = markers.size();
		if (size == 0)
			return null;
		currentMarker++;
		if (currentMarker >= size)
			currentMarker = 0;
		FileMarker marker = markers.get(currentMarker);
		marker.jump(view);
		selectInDockable(marker);
		return marker;
	}
	public FileMarker prevMarker(View view) {
		int size = markers.size();
		if (size == 0)
			return null;
		currentMarker--;
		if (currentMarker < 0 || currentMarker >= size)
			currentMarker = size - 1;
		FileMarker marker = markers.get(currentMarker);
		marker.jump(view);
		selectInDockable(marker);
		return marker;
	}

	private void selectInDockable(FileMarker marker) {
		View v = jEdit.getActiveView();
		if (v != null) {
			DockableWindowManager dwm = v.getDockableWindowManager();
			if (dwm != null) {
				JComponent c = dwm.getDockable("marker-set-manager");
				if (c != null) {
					MarkerSetManager msm = (MarkerSetManager) c;
					if (msm != null)
						msm.setSelected(marker);
				}
			}
		}	
	}

	/** @return true if the marker was added, false if removed */
	public boolean toggle(FileMarker marker) {
		if (markers.contains(marker))
		{
			remove(marker);
			return false;
		}
		add(marker);
		return true;
	}
	
	public void add(FileMarker marker) {
		if (markers.contains(marker))
			return;
		markers.add(marker);
		MarkerSetsPlugin.notifyChange(Event.MARKER_ADDED, marker, this);
		MarkerSetsPlugin.saveState();
		repaintAllTextAreas();
	}
	
	public void remove(FileMarker marker) {
		if (! markers.remove(marker))
			return;
		MarkerSetsPlugin.notifyChange(Event.MARKER_REMOVED, marker, this);
		MarkerSetsPlugin.saveState();
		repaintAllTextAreas();
	}
	
	public void handleBufferUpdate(BufferUpdate bu) {
		Object event = bu.getWhat();
		if ((event != BufferUpdate.LOADED) && (event != BufferUpdate.CLOSED))
			return;
		Buffer b = bu.getBuffer();
		Vector<FileMarker> pathMarkers = getMarkersFor(b.getPath());
		for (FileMarker marker: pathMarkers)
		{
			if (event == BufferUpdate.LOADED)
				marker.createPosition(b);
			else
				marker.removePosition();
		}
	}
	
	@SuppressWarnings("unchecked")
	public Vector<FileMarker> getMarkers() {
		Vector<FileMarker> items = new Vector<FileMarker>(markers);
		Collections.sort(items);
		return items;
	}
	
	public FileMarker getMarkerByShortcut(String shortcut)
	{
		for (int i = 0; i < markers.size(); i++)
		{
			FileMarker marker = markers.get(i);
			if (shortcut.equals(marker.shortcut))
				return marker;
		}
		return null;
	}
	
	public FileMarker getMarkerFor(String path, int line)
	{
		for (int i = 0; i < markers.size(); i++)
		{
			FileMarker marker = markers.get(i);
			if (marker.file.equals(path) && marker.getLine() == line)
				return marker;
		}
		return null;
	}

	// SideKick support: Is there a marker in specified position range?
	public boolean hasMarker(Buffer b, int start, int end)
	{
		for (FileMarker marker: markers)
		{
			if (marker.isInRange(b, start, end))
				return true;
		}
		return false;
	}

	public Vector<FileMarker> getMarkersFor(String path)
	{
		Vector<FileMarker> pathMarkers = new Vector<FileMarker>();
		for (int i = 0; i < markers.size(); i++)
		{
			FileMarker marker = markers.get(i);
			if (marker.file.equals(path))
				pathMarkers.add(marker);
		}
		return pathMarkers;
	}
	
	public void importXml(Element node)
	{
		name = node.getAttribute(NAME_ATTR);
		color = SyntaxUtilities.parseColor(node.getAttribute(COLOR_ATTR),
			Color.black);
		NodeList children = node.getChildNodes();
		for (int i = 0; i < children.getLength(); i++)
		{
			Node child = children.item(i);
			if (child instanceof Element)
				markers.add(new FileMarker((Element) child));
		}
	}
	public void exportXml(Element parent)
	{
		Document doc = parent.getOwnerDocument();
		Element setNode = doc.createElement("MarkerSet");
		parent.appendChild(setNode);
		setNode.setAttribute(NAME_ATTR, getName());
		setNode.setAttribute(COLOR_ATTR,
			SyntaxUtilities.getColorHexString(color));
		for (int i = 0; i < markers.size(); i++)
			markers.get(i).exportXml(setNode);
	}

	private void repaintAllTextAreas()
	{
		jEdit.visit(new JEditVisitorAdapter() {
			@Override
			public void visit(JEditTextArea textArea) {
				textArea.getGutter().repaint();
			}
		});
	}
}
